/***************************************************************************
                          ksmoothdock.cpp  -  description
                             -------------------
    begin                : May 7 22:18:34 BST 2004
    copyright            : (C) 2004 by KSmoothDock team
    email                : dangvd@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <alloc.h>

#include <qpixmap.h>
#include <qimage.h>
#include <qpainter.h>
#include <qcursor.h>
#include <qdir.h>
#include <qstringlist.h>
#include <qvaluelist.h>
#include <qrect.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qbrush.h>
#include <qpoint.h>

#include <kwin.h>
#include <kprocess.h>
#include <netwm_def.h>
#include <kmessagebox.h>
#include <klocale.h>

#include <dcopclient.h>
#include <qcstring.h> 
#include <qdatastream.h>

#include "ksmoothdock.h"
#include "showdesktop.h"
#include "welcomedialog.h"
#include "confdialog.h"
#include "confkickerdialog.h"
#include "aboutdialog.h"
#include "tooltip.h"

// macros

#define min(a, b) (a < b) ? (a) : (b)
#define max(a, b) (a > b) ? (a) : (b)

// constructor

KSmoothDock::KSmoothDock(QWidget *parent, const char *name) : QWidget(parent, name, WStyle_Customize |  WRepaintNoErase| WStyle_NoBorder | WDestructiveClose)
{
  int i;
	
	prepareWallpaper();

  // paths
  
  progPath = QDir::homeDirPath() + QString("/.ksmoothdock/");
  menuPath = progPath + QString("menu/");
  confFile = progPath + QString("ksmoothdock-3.3.conf");

  // init dynamic variables

  confDlg = NULL;
  aboutDlg = NULL;
  popup = NULL;
  info = NULL;
  wm = NULL;



  rootpix = NULL;

  animTimer = NULL;

  clockTimer = NULL;   
	
	toolTipWin = NULL;
  
  // load configuration 
  
  loadConf();
	
	// initiate window mode
	
	wMode = MINIMISED;
  
  // init icon sizes
  
  if (zoomMode == NORMAL_ZOOM) {
    iwSmall = iwSmallNZ;
    iwBig = iwBigNZ;
  } else {
    iwSmall = iwSmallPZ;
    iwBig = iwBigPZ;
  }

  // init spaces between icons and other drawing variables

  if (zoomMode == NORMAL_ZOOM) {
    iy = (iwBig - iwSmall)/2;
    ix = iy + 4;

    iSpace = iy + 2;

    iDist = iwSmall + iSpace;
  } else {
    ix = 10;

    iSpace = 10;

    iDist = iwSmall + iSpace;
    funcW = (int)(2.5*iDist);
  // funcW = iwSmall/2;
    funcH = iwBig - iwSmall;
    iy = (iDist - iwSmall) / 2;
  }

  // init NETWinInfo and set various aspects of the dock, e.g. showed on all desktops, skip pager, skip taskbar ...

  info = new NETWinInfo( qt_xdisplay(), winId(), qt_xrootwin(), NET::WMState);
  info->setDesktop( NETWinInfo::OnAllDesktops );
  info->setState(NETWinInfo::SkipTaskbar | NETWinInfo::SkipPager, NETWinInfo::SkipTaskbar | NETWinInfo::SkipPager);
  setBackgroundMode(NoBackground);

  // set items array (array of icons) to auto delete (i.e. all the items in the array will be deleted when the array is deleted


  items.setAutoDelete(TRUE);

  // init pager variables

  wm = new KWinModule();
  curDesk = wm->currentDesktop();
  numDesks = wm->numberOfDesktops();

  for (i = 0; i < numDesks; i++) {
    wallpapers[i] = NULL;
		fWallpapersReady[i] = false;
  }
	
  if (zoomMode == NORMAL_ZOOM) {
    for (i = 0; i < numDesks; i++) {
			tnWallpapers[i] = NULL;
    }
  }
	 
  // load the quick launch menu

  loadMenu();

  // add pager

  if (zoomMode == PARABOLIC_ZOOM) {
    if (fShowPager)
      addPager();
  }

  // load running tasks

  loadTasks();

  // connect the current desktop changed signal

  connect(wm, SIGNAL(currentDesktopChanged(int)), SLOT(currentDesktopChanged(int)));

  // transparency code

  rootpix = new KRootPixmap(this);
  rootpix->setCustomPainting(TRUE);
  connect(rootpix, SIGNAL(backgroundUpdated (const QPixmap &)), this, SLOT(updateBackground(const QPixmap &)));
  rootpix->start();
	hiddenBg = QPixmap(HIDDEN_WIDTH, MAX_HIDDEN_HEIGHT);

  // add main popup menu

  addMainPopupMenu();

  // add about dialog

  addAboutDlg();
	
	// tool tip stuff
	
	
	toolTipWin = new ToolTip();
	prevInd = curInd = -1;
	toolTipWin->setFontFace("Sans Serif");
	//toolTipWin->setItalic();
	//toolTipWin->setBold();
	toolTipWin->setSize(18);
	toolTipWin->setShadowOffset(0);
	toolTipWin->setColor(Qt::white);
	toolTipWin->setBgColor(QColor("#000088"));
	//toolTipWin->setBgColor(QColor("#0066cc"));	
	
	
	/*
	if (fShowToolTip) {
		toolTipWin = new ToolTip();
		prevInd = curInd = -1;
		toolTipWin->setFontFace(ttFontFace);
		toolTipWin->setItalic(ttIsItalic);
		toolTipWin->setBold(ttIsBold);
		toolTipWin->setSize(ttFontSize);
		toolTipWin->setShadowOffset(2);
		toolTipWin->setColor(ttColor);
		toolTipWin->setBgColor(ttBgColor);
	}
	*/
	

  // init animation variables in normal zoom mode

  if (zoomMode == NORMAL_ZOOM) {
    last_mx = -1;

    //  cout << getIndex(last_mx) << endl;
    animTimer = NULL;
    for (i = 0; i < MAX_ICONS; i++) {
      iSizeInc[i] = 0;
    }
  }

  // init desktop width

  if (scrWidth != 0) { // if using xinerama
    dw = scrWidth;
  } else {
    dw = QApplication::desktop()->width();
  }

  if (zoomMode == NORMAL_ZOOM) { // normal zoom mode

  	// init various variables

	  h = 2*iy + iwSmall;

	  dh = QApplication::desktop()->height();	  

	  // pager

	  if (fShowPager) {
	    pagerDeskH = (int)((iwSmall + 4) / 3) * 3; // make sure pagerDeskH is divisible by 3
						// so that the wallpaper thumbnails are not distorted
	    pagerDeskW = pagerDeskH * 4 / 3;
	    pagerW = (pagerDeskW + 2) * numDesks;
	  } else {
	    pagerW = 1; // we will draw a line to separate the quick launchers and the task bar
	  }

	  // clock

  	if (fShowClock) {
    	clockW = iDist + 8;
    	clockTimer = new QTimer(this);
    	connect( clockTimer, SIGNAL(timeout()), this, SLOT(clockTicked()) );
    	clockTimer->start( 60000, FALSE );
    	fUpdateClock = false;
  	} else {
  	  clockTimer = NULL;
  	  clockW = 0;
  	}

  	w = 2*ix + (items.count() - 1)*iDist + iwSmall + pagerW + clockW;
  	x = (dw - w) / 2;

  	for (i = 0; i < (int)items.count() + 1; i++) {
  	iSize[i] = iwSmall;
  	}

  	for (i = 0; i < numLaunchers; i++) {
  	  cx[i] = ix + i*iDist + iwSmall/2;
  	}
  	for (i = numLaunchers; i < (int)items.count() + 1; i++) {
  	  cx[i] = ix + i*iDist + iwSmall/2 + pagerW;
  	}

  	// init offscreen buffer

  	offscr = QPixmap(dw, h);

  	setMouseTracking(TRUE);

  	iOnClick = -1;
  	iOnChange = -1;
  	idOnChange = 0;
  	onChangeTimer = NULL;
  	onChangeAnimStep = 0;

  	ii_updating_first = 0;
  	ii_updating_last = items.count() - 1;
		
		loadWallpaper();

  	// set Strut of the dock

  	KWin::setStrut( winId(), 0, 0, 0, h );
		
  	// move and resize it

  	move( x, dh - h );
  	resize(w, h);
  } else { // PARABOLIC ZOOMING

  	// init various variables

	// clock

	if (fShowClock) {
	clockW = iDist + 16;
	clockTimer = new QTimer(this);
	connect( clockTimer, SIGNAL(timeout()), this, SLOT(clockTicked()) );
	clockTimer->start( 60000, FALSE );
	fUpdateClock = false;
	} else {
	clockTimer = NULL;
	clockW = 0;
	}

	for (i = 0; i < (int)items.count() + 1; i++) {
	iSize[i] = iwSmall;
	}

	for (i = 0; i < (int)items.count() + 1; i++) {
	cx[i] = ix + i*iDist + iwSmall/2;
	cur_cx[i] = cx[i];
	}

	w0 = 2*ix + ((int)items.count() - 1)*iDist + iwSmall;

	if (fShowTaskbar)
	w0 += SPACE_W;
	if (fShowClock)
	w0 += clockW;

	x0 = (dw - w0) / 2;
	h0 = 2*iy + iwSmall;
	h1 = 2*iy + iwBig;

	dh = QApplication::desktop()->height();

	setMouseTracking(TRUE);

	ii_first = 0;
	ii_last = (int)items.count() - 1;
	last_ii_first = 0;
	last_ii_last = (int)items.count() - 1;
	ii_updating_first = 0;
	ii_updating_last = ii_last;


	if (items.count() >= 5) {
	w1 = func(0) + 2*func(iDist) + 2*func(2*iDist) - 5*iwSmall + w0;
	} else if (items.count() == 4) {
	w1 = func(0) + 2*func(iDist) + func(2*iDist) - 4*iwSmall + w0;
	} else if (items.count() == 3) {
	w1 = func(0) + 2*func(iDist) - 3*iwSmall + w0;
	} else if (items.count() == 2) {
	w1 = func(0) + func(iDist) - 2*iwSmall + w0;
	} else if (items.count() == 1) {
	w1 = func(0) - iwSmall + w0;
	}

	x1 = (dw - w1)/2;
	deltaX = x0 - x1;
	deltaH = h1 - h0;
	sx = 0;
	sw = w1;

	fJustEntered = FALSE;
	//  fDrawAsJustEntered = FALSE;
	iOnClick = -1;
	iOnChange = -1;
	idOnChange = 0;
	onChangeTimer = NULL;
	onChangeAnimStep = 0;
	wMode = MINIMISED;

	/*
	isActive = FALSE;
	showDesktop();
	isActive = TRUE;



	*/


	offscr = QPixmap(dw, h1);
	bottomBg = QPixmap(dw, h0);
	bottomBgOrg = QPixmap(dw, h0);
	topBg = QPixmap(dw, deltaH);

	loadWallpaper();

	KWin::setStrut( winId(), 0, 0, 0, iDist );

	move( x1, dh - h1 );
	resize(w1, h1);
	setStateBelow();
  }
			
	setKickerPosition();
}

// destructor

KSmoothDock::~KSmoothDock()
{
  if (confDlg)
    delete confDlg;

  if (aboutDlg)
    delete aboutDlg;

  if (popup)
    delete popup;

  if (info)
    delete info;

  if (wm)
    delete wm;

  if (rootpix)
    delete rootpix;

  if (animTimer)
    delete animTimer;

  if (clockTimer)
    delete clockTimer;
		
	if (toolTipWin)
		delete toolTipWin;
}

// every second, we need to update the clock

void KSmoothDock::clockTicked() {
	if (wMode != HIDDEN) {
  	fUpdateClock = true;
  	repaint();
	}
}

// paint event, call the appropriate paint event handler for NORMAL ZOOM or PARABOLIC ZOOM

void KSmoothDock::paintEvent(QPaintEvent *e) {
	if (wMode == HIDDEN) {
		paintHiddenMode();
	} else {
  	if (zoomMode == NORMAL_ZOOM)
    	paintEventNZ(e);
  	else
    	paintEventPZ(e);
	}	
}

// paint in hidden mode

void KSmoothDock::paintHiddenMode() {
  QPainter p(&offscr);
	QPointArray array(3);
	
	array[0] = QPoint(hw/3, hh/2 - hw/3);
	array[1] = QPoint(hw/3, hh/2 + hw/3);
	array[2] = QPoint(2*hw/3, hh/2);
	
	bitBlt(&offscr, 0, 0, &hiddenBg, 0, 0, hw, hh);
	
	p.setBrush(borderColor);
	p.setPen(borderColor);
	//p.drawPolyline(array);
	p.drawPolygon(array);
	
	p.fillRect (QRect(0, 0, SPACE_W, hh), QBrush(borderColor));
	p.fillRect (QRect(0, 0, hw, SPACE_W), QBrush(borderColor));
	p.fillRect (QRect(hw - SPACE_W, 0, SPACE_W, hh), QBrush(borderColor));
	
	bitBlt(this, 0, 0, &offscr, 0, 0, hw, hh);
}

// paint event handler for Normal Zooming

void KSmoothDock::paintEventNZ(QPaintEvent *e)
{
  int i;
  int sx, sw;

  QPainter p(&offscr);

  // draw the clock only if fUpdateClock == TRUE

  if (fShowClock && fUpdateClock) {
    int rx;
    if ((int)items.count() > numLaunchers)
      rx = cx[items.count() - 1] + iwSmall/2;
    else
      rx = cx[items.count() - 1] + iwSmall/2 + iSpace + pagerW;

    QRect r(rx, 1, clockW + ix - 1, h - 1);
    bitBlt(&offscr, r.topLeft(), &bottomBg, r);

    /*
    QFont font = p.font();
    font.setPointSize(24
    );
    font.setBold(true);
    p.setFont(font);
    */

    QRect rClock(rx, 0, clockW, h);
		QFont font("Sans Serif", iwSmall/2);
		p.setFont(font);
    p.setPen(clockTextColor);
    p.drawText(rClock, AlignCenter, QTime::currentTime().toString("hh:mm"));

    /*
    font.setPointSize(12);
    p.setFont(font);
    */


    bitBlt(this, r.topLeft(), &offscr, r);
    fUpdateClock = false;
    return;
  }


//  cout << ii_updating_first << " " << ii_updating_last << endl;

  if (ii_updating_last != -1) {

    // draw the seperator

    /*
    if (fShowTaskbar) {
      QRect sepRect(cx[numLaunchers] - iwSmall/2 - iSpace/2, 0, pagerW, h);
      p.fillRect (sepRect, QBrush(sepColor));
    }
    */


    sx = cx[ii_updating_first] - iwSmall/2 - iSpace;
    sw = cx[ii_updating_last] - cx[ii_updating_first] + iwSmall + 2*iSpace;

    if (ii_updating_first == 0) {
      sx = 0;
      sw = cx[ii_updating_last] + iwSmall/2 + iSpace;
    }

    if (ii_updating_last == (int)items.count() - 1)
      sw = w + 1 - sx;

		if (fWallpapersReady[curDesk - 1]) {
    	bitBlt(&offscr, sx, 0, &bottomBg, sx, 0, sw, h);
		} else {
			prepareWallpaper();
			loadWallpaper();
		}


    if (fShowBorders) {
      p.fillRect (QRect(0, 0, SPACE_W, h), QBrush(borderColor));
      p.fillRect (QRect(0, 0, w, SPACE_W), QBrush(borderColor));
      p.fillRect (QRect(w - SPACE_W, 0, SPACE_W, h), QBrush(borderColor));
    }

    // draw the pager

    if (fShowPager) {
      for (i = 1; i <= numDesks; i++) {
        QRect r(cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 + (pagerDeskW + 2)*(i - 1), (h - pagerDeskH - 2)/2, pagerDeskW + 2, pagerDeskH + 2);

        if (fWallpapersReady[i - 1]) {
          if (tnWallpapers[i - 1] != NULL)
            bitBlt(&offscr, r.x() + 1, r.y() + 1, tnWallpapers[i - 1]);
        }


        if (i == curDesk)
          p.setPen(activeDeskColor);
        else
          p.setPen(inactiveDeskColor);
        p.drawRect(r);
        p.setPen(Qt::white);
        p.drawText(r, AlignCenter, QString::number(i));

      }
    } else if(fShowTaskbar) { // draw the separator
      QRect sepRect(cx[numLaunchers - 1] + iwSmall/2 + iSpace/2, 0, 1, h);
      p.fillRect (sepRect, QBrush(sepColor));
    }

    // draw icons

    Item *item;
    for ( i = ii_updating_first; i <= ii_updating_last; i++) {
      item = items.at(i);
      if (iOnClick == i) { // acknowledge when the user click on a quick launcher
        KPixmap *clickedIcon = new KPixmap(*item->getIcon(iSize[i]));
        KPixmapEffect::fade(*clickedIcon, 0.5, QColor("#FFFFFF"));
        bitBlt(&offscr, cx[i] - iSize[i]/2, (h - iSize[i])/2, clickedIcon);

		    QTimer::singleShot( 500, this, SLOT( stopTaskLauncherNotification() ) );

        delete clickedIcon;
      } else {
        QPixmap *pix = item->getIcon(iSize[i]);
        if (pix != NULL) {
          bitBlt(&offscr, cx[i] - iSize[i]/2, (h - iSize[i])/2, pix);
        }
      }
    }
  }

  // draw the notification animation

  if (fShowNotification) {
    if (iOnChange >= numLaunchers && iOnChange < (int)items.count()) {

      i = iOnChange;
      int tempx = cx[i] - iwSmall/2 - iSpace/2;
      int tempy = 1;
      int tempw = iwSmall + iSpace;
      int temph = h - 1;

      QRect sepRect(tempx, tempy, tempw, temph);
        bitBlt(&offscr, QPoint(tempx, tempy), &bottomBg, sepRect);

      QPixmap *pix = items.at(i)->getIcon(iwSmall);
      if (onChangeAnimStep == 1) {
        KPixmap *changedIcon = new KPixmap(*pix);
        KPixmapEffect::fade(*changedIcon, 0.5, QColor("#FFFFFF"));

        bitBlt(&offscr, cx[i] - iwSmall/2, (h - iwSmall)/2, changedIcon);

        delete changedIcon;
      } else {
        if (pix != NULL)
          bitBlt(&offscr, cx[i] - iwSmall/2, (h - iwSmall)/2, pix);
      }

      if (i < ii_updating_first || i > ii_updating_last) {
        bitBlt(this, cx[i] - iwSmall/2 - iSpace/2, 0,
        &offscr, cur_cx[i] - iwSmall/2 - iSpace/2, 0, iwSmall + iSpace, h);
      }
    }
  }

  // draw the clock

  if (fShowClock && ii_updating_last == (int)items.count() - 1) {
    int rx;
    if ((int)items.count() > numLaunchers)
      rx = cx[items.count() - 1] + iwSmall/2;
    else
      rx = cx[items.count() - 1] + iwSmall/2 + iSpace + pagerW;

    QRect rClock(rx, 0, clockW, h);
    p.setPen(clockTextColor);
		QFont font("Sans Serif", iwSmall/2);
		p.setFont(font);
    p.drawText(rClock, AlignCenter, QTime::currentTime().toString("hh:mm"));
  }

  // update the widget

  if (ii_updating_last != -1) {
    bitBlt(this, sx, 0, &offscr, sx, 0, sw, h);
  }
}

// paint event handler for Parabolic Zooming

void KSmoothDock::paintEventPZ(QPaintEvent *e)
{
  int i;

  QPainter p(&offscr);
  // draw the clock only if fUpdateClock == TRUE

  QFont defFont("Sans Serif", iwSmall/2);
  p.setFont(defFont);

  if (fShowClock && fUpdateClock) {
    int rx;

    if (wMode == MINIMISED) {
      rx = w1 - deltaX - clockW - ix/2;
    } else {
      rx = w1 - clockW - ix/2;
    }

    QRect r(rx + 1, deltaH + 1, clockW + ix/2 - 3, h0 - 2);
    QRect r2(rx + 1, 1, clockW + ix/2 - 3, h0 - 2);
    bitBlt(&offscr, r.topLeft(), &bottomBg, r2);

    /*
    QFont font = p.font();
    font.setPointSize(24
    );
    font.setBold(true);
    p.setFont(font);
    */

    QRect rClock(rx, deltaH, clockW, h0);

    p.setPen(clockTextColor);
    QFont font("Sans Serif", iwSmall/2);
    p.setFont(font);
    p.drawText(rClock, AlignCenter, QTime::currentTime().toString("hh:mm"));

    /*
    font.setPointSize(12);
    p.setFont(font);
    */


    bitBlt(this, r.topLeft(), &offscr, r);


    fUpdateClock = false;
    return;
  }

  if (wMode == MINIMISED) {
    //cout << "0" << deltaX << (w1 - deltaX) << w1 << endl;

    //offscr.fill(Qt::black);

    // bitBlt(&offscr, 0, 0, &bottomBgOrg, 0, 0, w1, deltaH);

    // fill the off screen buffer with bottom background - the transparent background of the dock

    // if (bottomBgOrg != NULL);

    if (fWallpapersReady[curDesk - 1]) {
    	bitBlt(&offscr, 0, deltaH, &bottomBgOrg, 0, 0, deltaX, h0);
    	bitBlt(&offscr, w1 - deltaX, deltaH, &bottomBgOrg, w1 - deltaX, 0, deltaX + 1, h0);
    	bitBlt(&offscr, deltaX, deltaH, &bottomBg, deltaX, 0, w1 - 2*deltaX, h0);

    	// fill top of the off screen buffer with real background - so that the dock looks "small"

    	bitBlt(&offscr, 0, 0, &topBg, 0, 0, w1 + 1 + iDist, deltaH); // + iDist: fix the bug: incorrect draw when new window added
	// cout << "Bit blt " << w1 << endl;
    } else {
			prepareWallpaper();
			loadWallpaper();
		}

    // draw borders

    if (fShowBorders) {
      p.fillRect (QRect(deltaX, deltaH, SPACE_W, h0), QBrush(borderColor));
      p.fillRect (QRect(deltaX, deltaH, w0, SPACE_W), QBrush(borderColor));
      p.fillRect (QRect(w1 - SPACE_W - deltaX, deltaH, SPACE_W, h0), QBrush(borderColor));
    }

    // draw items

    Item *item;
    i = 0;
    for (i = 0, item = items.first(); item; item = items.next(), i++) {
      if (i < numLaunchers) {
        bitBlt(&offscr, cx[i] - iwSmall/2 + deltaX, h1 - iy - iwSmall, item->getIcon(iwSmall));
      } else {
        bitBlt(&offscr, cx[i] - iwSmall/2 + SPACE_W + deltaX, h1 - iy - iwSmall, item->getIcon(iwSmall));
      }

      if (fShowPager && i >= numLaunchers && i < numLaunchers + numDesks) {
        int deskNum = i + 1 - numLaunchers;
	QRect r(cx[i] - iwSmall/2 + deltaX, h1 - iy - iwSmall, iwSmall, iwSmall*3/4);

	QFont font("Sans Serif", iwSmall/2);
	p.setFont(font);

	if (deskNum == curDesk)
		p.setPen(activeDeskColor);

	else
		p.setPen(inactiveDeskColor);


        p.drawText(r, AlignCenter, QString::number(deskNum));
      }
    }

    // draw the seperator when in minimised mode

    if (fShowPager || fShowTaskbar) {
      QRect sepRect(cx[numLaunchers] - iwSmall/2 - iSpace/2 + deltaX, deltaH, SPACE_W, h0);
      p.fillRect (sepRect, QBrush(sepColor));
    }

    if (fShowPager && fShowTaskbar) {
      QRect sepRect(cx[numLaunchers + numDesks] - iwSmall/2 - iSpace/2 + deltaX + SPACE_W, deltaH, SPACE_W, h0);
      p.fillRect (sepRect, QBrush(sepColor));
    }
		
		
    if (fShowClock) {
      QRect sepRect(w1 - deltaX - clockW - ix/2, deltaH, SPACE_W, h0);
      p.fillRect (sepRect, QBrush(sepColor));
    }
		

    // draw the notification animation

    if (fShowNotification) {
      if (iOnChange >= numLaunchers && iOnChange < (int)items.count()) {

        i = iOnChange;
        int tempx = cx[i] - iwSmall/2 - iSpace/2 + SPACE_W + deltaX;
        int tempy = deltaH + 1;
        int tempw = iwSmall + iSpace;
        int temph = iwSmall + iSpace - 1;

        QRect sepRect(tempx, tempy, tempw, temph);
        bitBlt(&offscr, QPoint(tempx, tempy),
            &bottomBg, sepRect);

        QPixmap *pix = items.at(i)->getIcon(iwSmall);

        if (onChangeAnimStep == 1) {
          KPixmap *changedIcon = new KPixmap(*pix);
          KPixmapEffect::fade(*changedIcon, 0.5, QColor("#FFFFFF"));

          bitBlt(&offscr, cx[i] - iSize[i]/2 + SPACE_W + deltaX, h1 - iy - iwSmall, changedIcon);

          delete changedIcon;
        } else {
          bitBlt(&offscr, cx[i] - iSize[i]/2 + SPACE_W + deltaX, h1 - iy - iwSmall, pix);
        }

        if (i < ii_updating_first || i > ii_updating_last) {
          bitBlt(this, cx[i] - iwSmall/2 - iSpace/2 + SPACE_W + deltaX, deltaH,
          &offscr, cx[i] - iwSmall/2 - iSpace/2 + SPACE_W + deltaX, 0, iwSmall + iSpace, h0);
        }
      }
    }

    if (fShowClock) {
			int rx;
			rx = w1 - deltaX - clockW - ix/2;
		
		
			QRect rClock(rx, deltaH, clockW, h0);
			QFont font("Sans Serif", iwSmall/2);
			p.setFont(font);
			p.setPen(clockTextColor);
			p.drawText(rClock, AlignCenter, QTime::currentTime().toString("hh:mm"));
    }

    // update the widget

    bitBlt(this, 0, 0, &offscr, 0, 0, w1 + 1 + iDist, h1); // + iDist: fix the bug: incorrect draw when new window added

  } else { // MAXIMISED
		if (fWallpapersReady[curDesk - 1]) {
			if (ii_updating_first == 0) {
	
				bitBlt(&offscr, 0, 0, &topBg, 0, 0, sw + sx + adjust + 1, deltaH);
				bitBlt(&offscr, 0, deltaH, &bottomBg, 0, 0, sw + sx + adjust, h0);
			} else {
				bitBlt(&offscr, sx + adjust, 0, &topBg, sx + adjust, 0, sw + 1, deltaH);
				bitBlt(&offscr, sx + adjust, deltaH, &bottomBg, sx + adjust, 0, sw, h0);
			}
		} else {
			prepareWallpaper();
			loadWallpaper();
		}

    if (fShowBorders) {
      p.fillRect (QRect(0, deltaH, SPACE_W, h0), QBrush(borderColor));
      p.fillRect (QRect(0, deltaH, w1, SPACE_W), QBrush(borderColor));
      p.fillRect (QRect(w1 - SPACE_W, deltaH, SPACE_W, h0), QBrush(borderColor));
    }


    Item *item;
    i = 0;
    for ( i = ii_updating_first; i <= ii_updating_last; i++) {
      item = items.at(i);

      if (i == iOnClick) { // acknowledge when the user click on a quick launcher
        KPixmap *clickedIcon = new KPixmap(*item->getIcon(iSize[i]));
        KPixmapEffect::fade(*clickedIcon, 0.5, QColor("#FFFFFF"));
        bitBlt(&offscr, cur_cx[i] - iSize[i]/2 + adjust, h1 - iy - iSize[i], clickedIcon);

		    QTimer::singleShot( 500, this, SLOT( stopTaskLauncherNotification() ) );

        delete clickedIcon;
      } else {
        bitBlt(&offscr, cur_cx[i] - iSize[i]/2 + adjust, h1 - iy - iSize[i], item->getIcon(iSize[i]));
      }

      if (fShowPager && i >= numLaunchers && i < numLaunchers + numDesks) {
        int deskNum = i + 1 - numLaunchers;
	QRect r(cur_cx[i] - iSize[i]/2 + adjust, h1 - iy - iSize[i], iSize[i], iSize[i]*3/4);

	QFont font("Sans Serif", iSize[i]/2);
	p.setFont(font);

	if (deskNum == curDesk)
		p.setPen(activeDeskColor);
	else
		p.setPen(inactiveDeskColor);

        p.drawText(r, AlignCenter, QString::number(deskNum));
      }
    }

    // draw the notification animation

    if (fShowNotification) {
      if (iOnChange >= numLaunchers && iOnChange < (int)items.count()) {
        i = iOnChange;
        int tempx = cur_cx[i] - iSize[i]/2 - iSpace/2 + adjust;
        int tempy = 1;
        int tempw = iSize[i] + iSpace;
        //int temph = iSize[i] + iSpace - tempy;

        QRect sepRect3(tempx, deltaH, tempw, h0);
        bitBlt(&offscr, QPoint(tempx, deltaH), &bottomBg, sepRect3);
        QRect sepRect4(tempx, tempy, tempw, deltaH - tempy);
        bitBlt(&offscr, QPoint(tempx, tempy), &topBg, sepRect4);

        QPixmap *pix = items.at(i)->getIcon(iSize[i]);

        if (onChangeAnimStep == 1) {
          KPixmap *changedIcon = new KPixmap(*pix);
          KPixmapEffect::fade(*changedIcon, 0.5, QColor("#FFFFFF"));

          bitBlt(&offscr, cur_cx[i] - iSize[i]/2 + adjust, h1 - iy - iSize[i], changedIcon);

          delete changedIcon;
        } else {
          bitBlt(&offscr, cur_cx[i] - iSize[i]/2 + adjust, h1 - iy - iSize[i], pix);
        }

        if (i < ii_updating_first || i > ii_updating_last) {
          bitBlt(this, cur_cx[i] - iSize[i]/2 - iSpace/2 + adjust, 0,
          &offscr, cur_cx[i] - iSize[i]/2 - iSpace/2 + adjust, 0, iSize[i] + iSpace, h1);
        }
      }
    }

    if (fShowClock && ii_updating_last >= (int)items.count() - 1) {
	int rx;
	rx = w1 - clockW - ix;

	QRect rClock(rx, deltaH, clockW, h0);
	QFont font("Sans Serif", iwSmall/2);
	p.setFont(font);

	p.setPen(clockTextColor);
	p.drawText(rClock, AlignCenter, QTime::currentTime().toString("hh:mm"));
    }

    // update the widget

    if (ii_updating_first == 0) {
      bitBlt(this, 0, 0, &offscr, 0, 0, sw + sx + adjust, h1);
    } else {
      bitBlt(this, sx + adjust, 0, &offscr, sx + adjust, 0, sw, h1);
    }
  }
}

/*
void KSmoothDock::mousePressEvent( QMouseEvent *e) {
  int i;

  int mx = e->x();
  int my = e->y();
  ButtonState srcButton = e->button();

  if (mx >= 1 && mx < cx[numLaunchers - 1] + iwSmall/2 + iSpace/2) { // Quick launcher
    i = getIndex(mx);

    if (srcButton == Qt::LeftButton) {
      if ((my >= 1) && (my <= h1 - 1)) {
      // launch the app
      // acknowledge on click

        iSize[i] = iwSmall;
        if (ii_updating_last == -1 || last_mx == -1) {
          ii_updating_first = i;
          ii_updating_last = i;
        } else {
          ii_updating_first = min(ii_updating_first, i);
          ii_updating_last = max(ii_updating_last, i);
        }
        repaint();

        if (i == 0) { // user clicked on the KMenu
          ii_updating_first = 0;
          ii_updating_last = items.count() - 1;
        }

        // run the app

        if (items.at(i)->getCommand().compare("SHOW_DESKTOP") == 0)
          showDesktop();
        else
          run(items.at(i)->getCommand());
      }
    } else if (srcButton == Qt::RightButton) {
      popup->exec(QCursor::pos());
      ii_updating_first = 0;
      ii_updating_last = items.count() - 1;
    }
  } else if (mx >= cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 && mx < cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 + pagerW) {
    if (srcButton == Qt::LeftButton) {
      // pager


      i = 1 + (int)(mx - (cx[numLaunchers - 1] + iwSmall/2 + iSpace/2)) / (pagerDeskW + 2);
      rootinfo = new NETRootInfo(qt_xdisplay(), 0xffffffff);
      rootinfo->setCurrentDesktop(i);
      delete rootinfo;
    }
  } else if (mx >= cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 + pagerW && mx <= w - clockW - ix - 1) {
    i = getIndex(mx);

    if (srcButton == Qt::LeftButton) {
      if ((my >= 1) && (my <= h - 1)) {
        WId id = items.at(i)->getId();
        if (id != activeId)
          KWin::setActiveWindow(id);
        else
          KWin::iconifyWindow(id, true);
      }
    } else if (srcButton == Qt::RightButton) {
      // user clicked on a running task icon, show the app menu.
      // pop-up menu for right-clicking on a taskbar icon.

      appMenu = new KPopupMenu(this);

      cId = items.at(i)->getId();
      appMenu->setTitle(KWin::info(cId).visibleName);
      appMenu->insertItem(i18n("Minimize"), this, SLOT(minApp()));
      appMenu->insertItem(i18n("Maximize"), this, SLOT(maxApp()));
      appMenu->insertItem(i18n("Restore"), this, SLOT(restApp()));
      appMenu->insertSeparator();
      appMenu->insertItem(i18n("Close"), this, SLOT(closeApp()));
      appMenu->exec(QCursor::pos());
      delete appMenu;

      ii_updating_first = 0;
      ii_updating_last = items.count() - 1;
    }
  }
}
*/

// mouse event handler

void KSmoothDock::mousePressEvent( QMouseEvent *e) {
	if (wMode == HIDDEN)
		doShow();
	else {
  	if (zoomMode == NORMAL_ZOOM)
    	mousePressEventNZ(e);
  	else
    	mousePressEventPZ(e);
	}
}

// mouse event handler for Normal Zooming

void KSmoothDock::mousePressEventNZ( QMouseEvent *e) {
  int i;

  int mx = e->x();
  int my = e->y();
  ButtonState srcButton = e->button();

  if (mx >= 1 && mx < cx[numLaunchers - 1] + iwSmall/2 + iSpace/2) { // Quick launcher
    i = getIndex(mx);

    if (srcButton == Qt::LeftButton) {
      if ((my >= 1) && (my <= h - 1)) {
      // launch the app
      // acknowledge on click

        if (items.at(i)->getCommand().compare("SHOW_DESKTOP") != 0) {
          iOnClick = i;
          if (ii_updating_last == -1 || last_mx == -1) {
            ii_updating_first = i;
            ii_updating_last = i;
          } else {
            ii_updating_first = min(ii_updating_first, i);
            ii_updating_last = max(ii_updating_last, i);
          }
          repaint();
        }

        if (i == 0) { // user clicked on the KMenu
          ii_updating_first = 0;
          ii_updating_last = (int)items.count() - 1;
        }

        // run the app

        if (items.at(i)->getCommand().compare("SHOW_DESKTOP") == 0) {
          showDesktop();
        } else {
          run(items.at(i)->getCommand());
        }
      }
    } else if (srcButton == Qt::RightButton) {
      popup->exec(QCursor::pos());
      ii_updating_first = 0;
      ii_updating_last = (int)items.count() - 1;
    }
  } else if (mx >= cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 && mx < cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 + pagerW) {
    if (srcButton == Qt::LeftButton) {
      // pager

      i = 1 + (int)(mx - (cx[numLaunchers - 1] + iwSmall/2 + iSpace/2)) / (pagerDeskW + 2);
      rootinfo = new NETRootInfo(qt_xdisplay(), 0xffffffff);
      rootinfo->setCurrentDesktop(i);
      delete rootinfo;
    }
  } else if (mx >= cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 + pagerW && mx <= w - clockW - ix - 1) {
    i = getIndex(mx);

    if (srcButton == Qt::LeftButton) {
      if ((my >= 1) && (my <= h - 1)) {
        WId id = items.at(i)->getId();
        if (id != activeId)
          KWin::setActiveWindow(id);
        else
          KWin::iconifyWindow(id, true);
      }
    } else if (srcButton == Qt::RightButton) {
      // user clicked on a running task icon, show the app menu.
      // pop-up menu for right-clicking on a taskbar icon.

      appMenu = new KPopupMenu(this);
      cId = items.at(i)->getId();
      appMenu->setTitle(KWin::info(cId).visibleName);
      appMenu->insertItem(i18n("Mi&nimize"), this, SLOT(minApp()));
      appMenu->insertItem(i18n("Ma&ximize"), this, SLOT(maxApp()));
      appMenu->insertItem(i18n("&Restore"), this, SLOT(restApp()));
      appMenu->insertSeparator();
      appMenu->insertItem(i18n("&Close"), this, SLOT(closeApp()));
      appMenu->exec(QCursor::pos());
      delete appMenu;

      ii_updating_first = 0;
      ii_updating_last = (int)items.count() - 1;
    }
  }
}

// mouse event handler for Parabolic Zooming

void KSmoothDock::mousePressEventPZ( QMouseEvent *e) {
  int i;

  int mx = e->x();
  int my = e->y();
  ButtonState srcButton = e->button();


  if (srcButton == Qt::LeftButton) {
    mx = mx + x1 - x0 - ix;

    i = mx/iDist;

    if ((my >= h1 - iy - iSize[i]) && (my <= h1 - iy)) {
      if ((i >= 0) && (i < (int)items.count())) {
        WId id = items.at(i)->getId();
        if (id == 0) { // if user clicked on a quick launcher, launch the app


          // run the app

          if (items.at(i)->getCommand().compare("SHOW_DESKTOP") == 0) {
            showDesktop();

            ii_updating_first = 0;
	          ii_updating_last = (int)items.count() - 1;
	          setStateBelow();
	          wMode = MINIMISED;
            fJustEntered = true;
	          updateBackground();
          } else if (items.at(i)->getCommand().startsWith("DESKTOP")) {
	          int deskNum = items.at(i)->getCommand().mid(7, 1).toInt();

	          NETRootInfo *rootinfo = new NETRootInfo(qt_xdisplay(), 0xffffffff);
            rootinfo->setCurrentDesktop(deskNum);
            delete rootinfo;

	          ii_updating_first = 0;
	          ii_updating_last = (int)items.count() - 1;
	          setStateBelow();
	          wMode = MINIMISED;
            fJustEntered = true;
	    // updateBackground();
	        } else {
            	// acknowledge on click
            	iOnClick = i;
            	repaint();

            	run(items.at(i)->getCommand()); 
	        }
        } else { // if user clicked on a running task's icon, activate/iconify it
          WId id = items.at(i)->getId();
          if (id != activeId)
            KWin::setActiveWindow(id);
          else {
            KWin::iconifyWindow(id, true);
					}
					setStateBelow();
	  			wMode = MINIMISED;
    			fJustEntered = true;
	  			updateBackground();
        }
      }
    }
  } else if (srcButton == Qt::RightButton) {
    mx = mx + x1 - x0 - ix;
    i = mx/iDist;

    if ((my >= h1 - iy - iSize[i]) && (my <= h1 - iy)) {
      if ((i >= 0) && (i < (int)items.count())) {

        WId id = items.at(i)->getId();
        if (id != 0) {
        // user clicked on a running task icon, show the app menu.
        // pop-up menu for right-clicking on a taskbar icon.

          appMenu = new KPopupMenu(this);
          cId = items.at(i)->getId();
          appMenu->setTitle(KWin::info(cId).visibleName);
      	  appMenu->insertItem(i18n("Mi&nimize"), this, SLOT(minApp()));
          appMenu->insertItem(i18n("Ma&ximize"), this, SLOT(maxApp()));
          appMenu->insertItem(i18n("&Restore"), this, SLOT(restApp()));
          appMenu->insertSeparator();
          appMenu->insertItem(i18n("&Close"), this, SLOT(closeApp()));
          appMenu->exec(QCursor::pos());
          delete appMenu;
	} else if (e->x() >= deltaX && e->x() <= w1 - deltaX) {
	  popup->exec(QCursor::pos());
	}
      }
    } else if (e->x() >= deltaX && e->x() <= w1 - deltaX) {
      popup->exec(QCursor::pos());
    }
  }
}

// mouse move event handler

void KSmoothDock::mouseMoveEvent( QMouseEvent *e) {
	if (wMode == HIDDEN) {
		mouseMoveEventHiddenMode(e);
	} else {
  	if (zoomMode == NORMAL_ZOOM)
    	mouseMoveEventNZ(e);
  	else
    	mouseMoveEventPZ(e);
	}
}

void KSmoothDock::mouseMoveEventHiddenMode(QMouseEvent *e) {
	curInd = 0;
	if (curInd != prevInd) {
		toolTipWin->setText("Show KSmoothDock");
		toolTipWin->move(1, dh - hh - toolTipWin->h - TOOLTIP_SPACE);
		if (!toolTipWin->isShown()) {
			toolTipWin->show();
		}
		prevInd = curInd;
	}	
}

// mouse move event handler for Normal Zooming

void KSmoothDock::mouseMoveEventNZ( QMouseEvent *e) {
  int i_prev, i_cur; // previous icon index, current icon index
  int mx = e->x();
  int my = e->y();
  int i;

  if ((my >= 1) && (my <= h - 1)) {

    if (mx >= cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 && mx < cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 + pagerW) { // on pager      
      last_mx = -1;
      ii_updating_first = 0;
      ii_updating_last = (int)items.count() - 1;

      for (i = 0; i < (int)items.count(); i++) {
        if (iSize[i] > iwSmall) {
          iSizeInc[i] = -1;
        }
      }

      if (animTimer == NULL) {
        animTimer = new QTimer(this);
        connect( animTimer, SIGNAL(timeout()), this, SLOT(timerTicked()) );
        animTimer->start( ANIM_STEP_TIME, FALSE );
      }
			
			// tooltip
			
			mx = mx - (cx[numLaunchers - 1] + iwSmall/2 + iSpace/2);
			
			int deskNum = mx/(pagerDeskW + 2); 
			curInd = numLaunchers + deskNum;
			
			if (curInd != prevInd) {
				toolTipWin->setText(QString("Desktop ") + QString::number(deskNum+1));
				toolTipWin->move(x + cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 - toolTipWin->w/2 + deskNum * pagerDeskW + pagerDeskW/2, dh - h - toolTipWin->h - TOOLTIP_SPACE);
			}
			
			if (!toolTipWin->isShown()) {
				toolTipWin->show();
			}
			
			prevInd = curInd;

      return; // quit
    }

    i_cur = getIndex(mx);

    if (last_mx == -1)
      i_prev = -1;
    else
      i_prev = getIndex(last_mx);

  //  cout << i_prev << " " << i_cur << endl;

    if (i_cur != i_prev) {      
      if (i_cur >= 0 && i_cur < (int)items.count()) {
        iSizeInc[i_cur] = 1;
				
				// tool tip
				
				if (i_cur < numLaunchers)
					curInd = i_cur;
				else
					curInd = i_cur + numDesks;
        
				if (curInd != prevInd) {
					toolTipWin->setText(items.at(i_cur)->getName());
					toolTipWin->move(x + cx[i_cur] - toolTipWin->w/2, dh - h - toolTipWin->h - TOOLTIP_SPACE);
				}
			
				if (!toolTipWin->isShown()) {
					toolTipWin->show();
				}
			
				prevInd = curInd;
				
				// end tool tip

  //      if ((fShowNotification) && (i_cur == iOnChange)) {
        if (fShowNotification) {
          iOnChange = -1;
          idOnChange = 0;

          if (onChangeTimer != NULL) {
            delete onChangeTimer;
            onChangeTimer = NULL;
          }
        }

        if (ii_updating_last == -1 || i_prev == -1) {
          ii_updating_first = i_cur;
          ii_updating_last = i_cur;
        } else {
          ii_updating_first = min(ii_updating_first, i_cur);
          ii_updating_last = max(ii_updating_last, i_cur);
        }
      } else if (i_cur == (int)items.count()) {
				// tool tip
				curInd = i_cur + numDesks;
				if (curInd != prevInd) {
					toolTipWin->setText(QDate::currentDate().toString("ddd dd MMM yyyy"));					
					toolTipWin->move(x + w - ix - clockW/2 - toolTipWin->w/2, dh - h - toolTipWin->h - TOOLTIP_SPACE);
					
					cout << x << " " << cx[items.count() - 1] << " " << toolTipWin->w << endl;
				}
			
				if (!toolTipWin->isShown()) {
					toolTipWin->show();
				}
			
				prevInd = curInd;
      }

//      cout << "i_cur=" << i_cur << ", i_prev=" << i_prev << ", items.count()=" << items.count() << endl;
      if (i_prev >= 0 && i_prev < (int)items.count()) {
        iSizeInc[i_prev] = -1;
        ii_updating_first = min(ii_updating_first, i_prev);
        ii_updating_last = max(ii_updating_last, i_prev);
      }
      if (animTimer == NULL) {
        animTimer = new QTimer(this);
        connect( animTimer, SIGNAL(timeout()), this, SLOT(timerTicked()) );
        animTimer->start( ANIM_STEP_TIME, FALSE );
      }
    }

    last_mx = e->x();
  }
}

// mouse move event handler for Parabolic Zooming

void KSmoothDock::mouseMoveEventPZ( QMouseEvent *e) {
  int i;
  int mx, my;
  int dx;
  int k;
  int cur_cx_desk;
	bool needToolTip = false;
	//bool needToolTipInit = false;

//  if (fDrawAsJustEntered) {

  if (fJustEntered && e->x() >= deltaX && e->x() <= w1 - deltaX) {
		needToolTip = true;
		//needToolTipInit = true;
		
    topBg = QPixmap::grabWindow(qt_xrootwin(), x1, dh - h1, w1, deltaH);
    setStateAbove();
    KWin::setActiveWindow(winId());

    mx = e->x() - deltaX;
    last_mx = e->x();

    for (i = 0; i < (int)items.count() + 1; i++) {
      dx = abs(cx[i] - mx);
      if (dx < funcW) {
        iSize[i] = func(dx);
      } else {
        iSize[i] = iwSmall;
      }
    }

    // debug

    // cout << mx << endl;
    //for (i = 0; i < items.count() + 1; i++) {
      //cout << abs(cx[i] - mx) << " " << iSize[i] << " ";
    //}
    //cout << endl;

    // end debug

    cur_cx[0] = ix + iSize[0]/2;

    for (i = 1; i < (int)items.count() + 1; i++) {
      cur_cx[i] = cur_cx[i-1] + (iSize[i] + iSize[i-1])/2 + iSpace;
    }

    for (k = 0; (k < (int)items.count()) && (mx > cx[k]); k++);

    cur_cx_desk = x0 + cx[k] + (int)((cx[k] - mx)*(func(0) + func(iDist) + 2*iSpace - 2*iDist)/(2*iDist));
    adjust = cur_cx_desk - cur_cx[k] - x1;

    ii_updating_first = 0;
    ii_updating_last = (int)items.count() - 1;
    last_ii_first = 0;
    last_ii_last = (int)items.count() - 1;
    ii_first = 0;
    ii_last = (int)items.count() - 1;
    sx = 0;
    sw = w1;


    fJustEntered = FALSE;
    //QApplication::syncX();
    wMode = MAXIMISED;
  } else if ((wMode == MAXIMISED) && (e->x() < deltaX || e->x() > w1 - deltaX)) { // process like on LeaveEvent
    leaveEvent(NULL);
    fJustEntered = TRUE;
  } else if ((wMode == MAXIMISED) && (abs(e->x() - last_mx) >= UPDATE_DIST)) {
		needToolTip = true;
		
    last_mx = e->x();

    last_ii_first = ii_first;
    last_ii_last = ii_last;

    ii_first = 0;
    ii_last = (int)items.count() - 1;


    mx = e->x() - deltaX;

    for (i = 0; i < (int)items.count() + 1; i++) {
      dx = abs(cx[i] - mx);
      if (dx < funcW) {
        iSize[i] = func(dx);
      } else {
        iSize[i] = iwSmall;

        if (i < (int)items.count()) {
          if (cx[i] < mx) {
            ii_first = min(i + 1, (int)items.count() - 1);
          } else if (ii_last == (int)items.count() - 1) {
            ii_last = max(i - 1, 0);
          }
        }
      }
    }

    cur_cx[0] = ix + iSize[0]/2;
    for (i = 1; i < (int)items.count() + 1; i++) {
      cur_cx[i] = cur_cx[i-1] + (iSize[i] + iSize[i-1])/2 + iSpace;
    }

    for (k = 0; (k < (int)items.count()) && (mx > cx[k]); k++);

    cur_cx_desk = x0 + cx[k] + (int)((cx[k] - mx)*(func(0) + func(iDist) + 2*iSpace - 2*iDist)/(2*iDist));
    adjust = cur_cx_desk - cur_cx[k] - x1;

    ii_updating_first = min(ii_first, last_ii_first);
    ii_updating_last = max(ii_last, last_ii_last);

    if (ii_updating_last < (int)items.count() - 1) {
      cur_cx[items.count() - 1] = w1 - iwSmall/2 - ix - adjust;
      for (i = (int)items.count() - 2; i >= ii_updating_last + 1; i--) {
        cur_cx[i] = cur_cx[i+1] - iDist;
      }

      if (ii_updating_last == 0)
        cur_cx[0] = cur_cx[1] - iDist - (iSize[0] - iSize[1])/2;
    }

    sx = cur_cx[ii_updating_first] - iSize[ii_updating_first]/2;
    sx -= min (iSpace/2, ix);
    sw = cur_cx[ii_updating_last] - sx + iSize[ii_updating_last]/2 + iSpace/2;


    if (ii_updating_first == 0) {
      sx = 0;
      sw = cur_cx[ii_updating_last] + iSize[ii_updating_last]/2 + iSpace/2;
    }

    if (ii_updating_last == (int)items.count() - 1)
      sw = w1 - sx;
  }	
	
  repaint();
	
	// tool tip  
	
	if (needToolTip) {
		cout << "Need tool tip" << endl;
		mx = e->x();
  	my = e->y();
  	mx = mx - deltaX - ix + iSpace/2;
		
		
		curInd = mx/iDist;
		cout << curInd << endl;
		if (curInd >= 0 && curInd < items.count() && (my >= h1 - iy - iSize[curInd]) && (my <= h1 - iy)) {
			if (curInd != prevInd) {
				toolTipWin->setText(items.at(curInd)->getName());
				//toolTipWin->setText("tt");
			}
		
			if (curInd == 0)
				toolTipWin->move(x1 + (cur_cx[curInd] - iwSmall/2 - iSpace/2 + adjust + 2*SPACE_W) - (toolTipWin->w/2), dh - h1 - toolTipWin->h);
			else
				toolTipWin->move(x1 + (cur_cx[curInd] - iwSmall/2 + adjust + 2*SPACE_W) - (toolTipWin->w/2), dh - h1 - toolTipWin->h);
			
			if (!toolTipWin->isShown()) {
				toolTipWin->show();
				//cout << "Show" << endl;
			}
			prevInd = curInd;
		} else if (curInd >= items.count()) { // clock
			cout << e->x() << " " << (w1 - clockW - deltaX) << endl;
			curInd = items.count();
			if (curInd != prevInd) {
				toolTipWin->setText(QDate::currentDate().toString("ddd dd MMM yyyy"));
				toolTipWin->move(x1 + w1 - ix - clockW/2 - toolTipWin->w/2, dh - h1 - toolTipWin->h);
			
				if (!toolTipWin->isShown())
					toolTipWin->show();
			}
			prevInd = curInd;
		} else {
			if (toolTipWin->isShown())
				toolTipWin->hide();
		}
	}
}

// enter event handler

void KSmoothDock::enterEvent(QEvent *e) {
	if (wMode == HIDDEN) {
		return;
	}

  if (zoomMode == NORMAL_ZOOM)
    enterEventNZ(e);
  else
    enterEventPZ(e);
}

void KSmoothDock::enterEventNZ(QEvent *e) {
  KWin::setActiveWindow(winId());
}

void KSmoothDock::enterEventPZ(QEvent *e) {
//  enterTimer = new QTimer(this);
//  connect( enterTimer, SIGNAL(timeout()), this, SLOT(enterTimerTicked()) );
//  enterTimer->start( ENTER_TIME, TRUE );

  // topBg = QPixmap::grabWindow(qt_xrootwin(), x1, dh - h1, w1, deltaH);
  // setStateAbove();
  // KWin::setActiveWindow(winId());

  fJustEntered = TRUE;
//  fDrawAsJustEntered = TRUE;

}

// leave event handler

void KSmoothDock::leaveEvent(QEvent *e) {
	// tool tip

	toolTipWin->hide();
	prevInd = curInd = -1;

	// return if in hidden mode
	
	if (wMode == HIDDEN) {
		return;
	}	
	
  if (zoomMode == NORMAL_ZOOM)
    leaveEventNZ(e);
  else
    leaveEventPZ(e);
}

void KSmoothDock::leaveEventNZ(QEvent *e) {
  int i;

  if (animTimer != NULL)
    delete animTimer;


  last_mx = -1;
  ii_updating_first = 0;
  ii_updating_last = (int)items.count() - 1;

  for (i = 0; i < (int)items.count(); i++) {
    if (iSize[i] > iwSmall || iSizeInc[i] > 0) {
      iSizeInc[i] = -1;
    }
  }

  animTimer = new QTimer(this);
  connect( animTimer, SIGNAL(timeout()), this, SLOT(timerTicked()) );
  animTimer->start( ANIM_STEP_TIME, FALSE );
}


void KSmoothDock::leaveEventPZ(QEvent *e) {	
	setStateBelow();

// update the top background

	if (wallpapers[curDesk - 1] != NULL) {
    QPixmap *pix = wallpapers[curDesk - 1];
    bitBlt(&topBg, 0, 0, pix, x1, dh - h1, w1 + 1 + iDist, deltaH);
	}
	
	// repaint in minimised mode
	
  wMode = MINIMISED;  
  repaint();
}

// call when the user click on "Edit Quick Launch Menu"

void KSmoothDock::edit() {
  KMessageBox::information(this, "Edit the quick launch menu by adding/removing Link To Application(s) to the menu directory. You will need to restart the program for the changes to take effect.\n\nNOTE: Quick Launch menu items are sorted alphabetically by the corresponding Link To Application's file names.");
  run(QString("konqueror ") + menuPath);
}

// call when the user click on "Edit Preferences"

void KSmoothDock::editPref() {
  //run(QString("kwrite ") + confFile);
  if (!confDlg) {
    confDlg = new ConfDialog(this, "KSmoothDock Preferences");
  }
  confDlg->updateValues();
  confDlg->show();
}

// run a specific command, e.g. start a program from the quick launch menu

void KSmoothDock::run(QString cmd) {
	int len;
	
	// remove leading and trailing single and double quotes
	
	len = cmd.length();
	if ((cmd.at(0) == '\'' && cmd.at(len - 1) == '\'') || (cmd.at(0) == '\"' && cmd.at(len - 1) == '\"')) {
		cmd = cmd.mid(1, len - 2);
	}
	
	// run the command

  KProcess *proc = new KProcess;
  QStringList list;
  list = QStringList::split(" ", cmd);
  for ( QStringList::Iterator it = list.begin(); it != list.end(); ++it ) {
    *proc << (*it);
  }

  proc->start(KProcess::DontCare);
//  cout << "Run command: " << cmd << endl;
}

// Parabolic function
// Input: the distant from the icon to the mouse position
// Output: icon size

int KSmoothDock::func(int x) {
  int res = iwBig - (int)((x*x*funcH)/(funcW*funcW));
  // int res = iwBig;
  return res;
}

// window added event handler

void KSmoothDock::windowAdded(WId id) {
	if (!fShowTaskbar || wMode == HIDDEN)
		return;

  if (!fShowWinAllDesks && !KWin::info(id).onAllDesktops && KWin::info(id).desktop != curDesk)
    return;

  if (id != winId()) {
    // cout << "Window added" << endl;
    addTask(id);
    doUpdateGeometry();
  }
}

// window removed event handler

void KSmoothDock::windowRemoved(WId id) {
  Item *item;
  int i;
	
	if (!fShowTaskbar || wMode == HIDDEN)
		return;	
	
  for ( i = numLaunchers; i < (int)items.count(); i++ ) {
    item = items.at(i);
    if (item->getId() == id) {
      items.remove();
      if (idOnChange == id) {
        idOnChange = 0;
        iOnChange = -1;
      } else if (i < iOnChange) {
        iOnChange--;
      }
      break;
    }
  }
  doUpdateGeometry();
}

// window changed event handler

void KSmoothDock::windowChanged(WId id, unsigned int properties) {
  Item *item;
  bool iconChanged, nameChanged, deskChanged;
  int i;
  bool toRepaint = FALSE;

	if (!fShowTaskbar || wMode == HIDDEN)
		return;	
	
  iconChanged = (properties & NET::WMIcon);
  nameChanged = (properties & NET::WMName);
	deskChanged = (properties & NET::WMDesktop);


  if (!fShowWinAllDesks && !KWin::info(id).onAllDesktops && KWin::info(id).desktop != curDesk && !deskChanged)
    return;
	if (deskChanged && fShowTaskbar && !fShowWinAllDesks) {
		if (KWin::info(id).desktop == curDesk) { // a window has been moved to the current desktop, add it
			windowAdded(id);
		} else { // a window has been moved to a different desktop, remove it
			windowRemoved(id);
		}
	}

  if (id != winId()) {
    for (i = numLaunchers, item = items.at(numLaunchers); item; i++, item = items.next()) {
      if (item->getId() == id) {
        if (nameChanged) {
          // Window title changed.
          item->setName(KWin::info(id).name);

          if ((fShowNotification) && (id != wm->activeWindow()) && (!KWin::info(id).name.startsWith("XMMS"))) {
            // An inactive window's title changed
            // notify user
            // ignore XMMS

            iOnChange = i;
            idOnChange = id;
            onChangeAnimStep = 0;

            if (onChangeTimer == NULL) {
              onChangeTimer = new QTimer(this);
              connect( onChangeTimer, SIGNAL(timeout()), this,
                SLOT(onChangeTimerTicked()) );
              onChangeTimer->start( ON_CHANGE_ANIM_INTERVAL, FALSE );
            }
            toRepaint = TRUE;
          }
        }

        if (iconChanged) {
          // Window icon changed.
          item->setIcon(KWin::icon(id, 64, 64));
          toRepaint = TRUE;
        }

        if (toRepaint)
          repaint();

        break;
      }
    }
  }
}

// current active window changed event handler

void KSmoothDock::activeWindowChanged(WId id) {
  bool toRepaint = FALSE;

	if (!fShowTaskbar || wMode == HIDDEN) {
		return;
	}
	
	activeId = id;

	/*
	if (isActive && id != winId()) {
		isActive = FALSE;
		wMode = MINIMISED;
		toRepaint = TRUE;
	}
	*/

	if ((fShowNotification) && (id == idOnChange)) {
		iOnChange = -1;
		idOnChange = 0;
		toRepaint = TRUE;
		if (onChangeTimer != NULL) {
			delete onChangeTimer;
			onChangeTimer = NULL;
		}
	}

	if (toRepaint) {
		repaint();
	}
}

// update geometry of the dock (e.g. width, height, layout ...)

void KSmoothDock::doUpdateGeometry() {
  if (zoomMode == NORMAL_ZOOM)
    doUpdateGeometryNZ();
  else
    doUpdateGeometryPZ();
}

void KSmoothDock::doUpdateGeometryNZ() {
  int i;

  iy = (iwBig - iwSmall)/2;
  ix = iy + 4;

  iSpace = iy + 2;

  iDist = iwSmall + iSpace;
  for (i = 0; i < numLaunchers; i++) {
    cx[i] = ix + i*iDist + iwSmall/2;
  }
  for (i = numLaunchers; i < (int)items.count() + 1; i++) {
    cx[i] = ix + i*iDist + iwSmall/2 + pagerW;
  }

  if (onChangeTimer != NULL) {
    delete onChangeTimer;
    onChangeTimer = NULL;
  }

  onChangeAnimStep = 0;

  if (animTimer != NULL) {
    delete animTimer;
    animTimer = NULL;
  }

  // init various variables

	h = 2*iy + iwSmall;

	dh = QApplication::desktop()->height();

	// pager

	if (fShowPager) {
	  pagerDeskH = (int)((iwSmall + 4) / 3) * 3; // make sure pagerDeskH is divisible by 3
						// so that the wallpaper thumbnails are not distorted
	  pagerDeskW = pagerDeskH * 4 / 3;
	  pagerW = (pagerDeskW + 2) * numDesks;
	} else {
	  pagerW = 1; // we will draw a line to separate the quick launchers and the task bar
	}

	// clock

	if (fShowClock) {
	  clockW = iDist + 8;
	  clockTimer = new QTimer(this);
	  connect( clockTimer, SIGNAL(timeout()), this, SLOT(clockTicked()) );
	  clockTimer->start( 60000, FALSE );
	  fUpdateClock = false;
	} else {
	  clockTimer = NULL;
	  clockW = 0;
	}

	w = 2*ix + ((int)items.count() - 1)*iDist + iwSmall + pagerW + clockW;

	x = (dw - w) / 2;

	for (i = 0; i < (int)items.count() + 1; i++) {
	  iSize[i] = iwSmall;
	}

	for (i = 0; i < numLaunchers; i++) {
	  cx[i] = ix + i*iDist + iwSmall/2;
	}
	for (i = numLaunchers; i < (int)items.count() + 1; i++) {
	  cx[i] = ix + i*iDist + iwSmall/2 + pagerW;
	}

	// init offscreen buffer

	offscr = QPixmap(dw, h);

	iOnClick = -1;
	iOnChange = -1;
	idOnChange = 0;
	onChangeTimer = NULL;
	onChangeAnimStep = 0;

	ii_updating_first = 0;
	ii_updating_last = (int)items.count() - 1;

	// set Strut of the dock

	KWin::setStrut( winId(), 0, 0, 0, h );

	// move and resize it

	move( x, dh - h );
	resize(w, h);

//  repaint();
}


void KSmoothDock::doUpdateGeometryPZ() {
  int i;

  ix = 10;

  iSpace = 10;

  iDist = iwSmall + iSpace;
  funcW = (int)(2.5*iDist);
  // funcW = iwSmall/2;
  funcH = iwBig - iwSmall;
  iy = (iDist - iwSmall) / 2;

  if (fShowClock) {
	  clockW = iDist + 16;
  } else {
    clockW = 0;
  }

  for (i = 0; i < (int)items.count() + 1; i++) {
    iSize[i] = iwSmall;
  }

  for (i = 0; i < (int)items.count() + 1; i++) {
    cx[i] = ix + i*iDist + iwSmall/2;
    cur_cx[i] = cx[i];
  }

  w0 = 2*ix + ((int)items.count() - 1)*iDist + iwSmall;

  if (fShowTaskbar)
    w0 += SPACE_W;
  if (fShowClock)
    w0 += clockW;

  x0 = (dw - w0) / 2;
  h0 = 2*iy + iwSmall;
  h1 = 2*iy + iwBig;

  dh = QApplication::desktop()->height();

  adjust = 0;

  ii_first = 0;
  ii_last = (int)items.count() - 1;
  last_ii_first = 0;
  last_ii_last = (int)items.count() - 1;
  ii_updating_first = 0;
  ii_updating_last = ii_last;


  if ((int)items.count() >= 5) {
    w1 = func(0) + 2*func(iDist) + 2*func(2*iDist) - 5*iwSmall + w0;
  } else if ((int)items.count() == 4) {
    w1 = func(0) + 2*func(iDist) + func(2*iDist) - 4*iwSmall + w0;
  } else if ((int)items.count() == 3) {
    w1 = func(0) + 2*func(iDist) - 3*iwSmall + w0;
  } else if ((int)items.count() == 2) {
    w1 = func(0) + func(iDist) - 2*iwSmall + w0;
  } else if ((int)items.count() == 1) {
    w1 = func(0) - iwSmall + w0;
  }

  x1 = (dw - w1)/2;
  deltaX = x0 - x1;

  deltaH = h1 - h0;
  sx = 0;
  sw = w1;

  fJustEntered = FALSE;
//  fDrawAsJustEntered = FALSE;
  iOnClick = -1;
  iOnChange = -1;
  idOnChange = 0;
  onChangeAnimStep = 0;
  wMode = MINIMISED;

  offscr = QPixmap(dw, h1);
	bottomBg = QPixmap(dw, h0);
	bottomBgOrg = QPixmap(dw, h0);
	topBg = QPixmap(dw, deltaH);

  move(x1, dh - h1);
  resize(w1, h1);
  KWin::setStrut( winId(), 0, 0, 0, iDist );

  setStateBelow();
  updateBackground();

  /*
  offscr = QPixmap(dw, h1);
  bottomBg = QPixmap(w1 + 1, h0);
  bottomBgOrg = QPixmap(w1 + 1, h0);
  topBg = QPixmap(w1 + 1, deltaH);
  */

  //repaint();
}

// add a running task to the items array


void KSmoothDock::addTask(WId id) {
  QPixmap pix = KWin::icon(id, 64, 64);
  KWin::Info info = KWin::info(id);
  QString name = info.name;

  if (!fShowWinAllDesks) {
     if (!info.onAllDesktops && info.desktop != curDesk) {
       return;
     }
  }
	
	//cout << info.windowType << endl;

  if (info.windowType == NET::Normal || info.windowType == NET::Override || info.windowType == NET::Unknown)
      if ((info.state & NET::SkipTaskbar) == 0)
        if (name.compare("XMMS Playlist") != 0 && name.compare("XMMS Equalizer") != 0) {
					// check if the item is already in the list
					
					for ( int i = numLaunchers; i < (int)items.count(); i++ ) {
    				Item *item = items.at(i);
    				if (item && item->getId() == id)
							return;
					}
					
					// if it is not already in the list, add it
				
          items.append(new Item(pix, id, name, iwSmall, iwBig));
//          cout << info.name << endl;
        }
}

// load configuration

void KSmoothDock::loadConf() {
  QString line, iconName;
  bool ok;
  QFile in(confFile);

  if (in.open(IO_ReadOnly)) {
    if (in.readLine(line, MAX_LEN) != -1) {
      if (line.contains("[KSmoothDock Configuration File]", FALSE) > 0) {
        while (in.readLine(line, MAX_LEN) != -1) {
	  if (line.startsWith("ZoomMode=")) {

            zoomMode = line.mid(9).stripWhiteSpace().toInt(&ok);
            if (!ok)
              zoomMode = NORMAL_ZOOM;
          } else if (line.startsWith("ShowPager=")) {
            fShowPager = line.mid(10).stripWhiteSpace().toInt(&ok);
            if (!ok)
              fShowPager = 1;
          } else if (line.startsWith("ShowTaskbar=")) {
            fShowTaskbar = line.mid(12).stripWhiteSpace().toInt(&ok);
            if (!ok)
              fShowTaskbar = 1;
          } else if (line.startsWith("ShowWindowsFromAllDesktops=")) {
            fShowWinAllDesks = line.mid(27).stripWhiteSpace().toInt(&ok);
            if (!ok)
              fShowWinAllDesks = 1;
          } else if (line.startsWith("ShowClock=")) {
            fShowClock = line.mid(10).stripWhiteSpace().toInt(&ok);
            if (!ok)
              fShowClock = 1;
          } else if (line.startsWith("ShowNotification=")) {
            fShowNotification = line.mid(17).stripWhiteSpace().toInt(&ok);

            if (!ok)
              fShowNotification = 0;
          } else if (line.startsWith("ShowBorders=")) {
             fShowBorders = line.mid(12).stripWhiteSpace().toInt(&ok);
             if (!ok)
               fShowBorders = 1;
          } else if (line.startsWith("BorderColor=")) {
            borderColor = QColor(line.mid(12).stripWhiteSpace());
          } else if (line.startsWith("SmallIconSizeNormalZoom=")) {
            iwSmallNZ = line.mid(24).stripWhiteSpace().toInt(&ok);
            if (!ok)
              iwSmallNZ = DEF_SMALL_SIZE_NZ;
          } else if (line.startsWith("BigIconSizeNormalZoom=")) {
            iwBigNZ = line.mid(22).stripWhiteSpace().toInt(&ok);
            if (!ok)
              iwBigNZ = DEF_BIG_SIZE_NZ;
          } else if (line.startsWith("SmallIconSizeParabolicZoom=")) {
            iwSmallPZ = line.mid(27).stripWhiteSpace().toInt(&ok);
            if (!ok)
              iwSmallPZ = DEF_SMALL_SIZE_PZ;
          } else if (line.startsWith("BigIconSizeParabolicZoom=")) {
            iwBigPZ = line.mid(25).stripWhiteSpace().toInt(&ok);
            if (!ok)
              iwBigPZ = DEF_BIG_SIZE_PZ;
          } else if (line.startsWith("DockOpacity=")) {
            dockOpacity = line.mid(12).stripWhiteSpace().toInt(&ok);
            if (!ok)
            dockOpacity = DEF_DOCK_OPACITY;
            if (dockOpacity < 0)
              dockOpacity = 0;
          else if (dockOpacity > 100)
              dockOpacity = 100;
          } else if (line.startsWith("BackgroundColor=")) {
            bgColor = QColor(line.mid(16).stripWhiteSpace());
          } else if (line.startsWith("SeparatorColor=")) {
            sepColor = QColor(line.mid(15).stripWhiteSpace());
          } else if (line.startsWith("ScreenWidth=")) {
            scrWidth = line.mid(12).stripWhiteSpace().toInt(&ok);
            if (!ok)
              scrWidth = 0;
          } else if (line.startsWith("ActiveDesktopColor=")) {
            activeDeskColor = QColor(line.mid(19).stripWhiteSpace());
          } else if (line.startsWith("InactiveDesktopColor=")) {
            inactiveDeskColor = QColor(line.mid(21).stripWhiteSpace());
          } else if (line.startsWith("ClockTextColor=")) {
            clockTextColor = QColor(line.mid(15).stripWhiteSpace());
	  			} 
					/*
					else if (line.startsWith("ShowToolTip=")) {
            fShowToolTip = line.mid(12).stripWhiteSpace().toInt(&ok);
	  			} else if (line.startsWith("ToolTipFontFace=")) {
            ttFontFace = line.mid(16).stripWhiteSpace();
	  			} else if (line.startsWith("ToolTipFontSize=")) {
            ttFontSize = line.mid(16).stripWhiteSpace().toInt(&ok);
	  			} else if (line.startsWith("ToolTipColor=")) {
            ttColor = QColor(line.mid(13).stripWhiteSpace());
	  			} */
        }

        if (iwSmallNZ > iwBigNZ || iwSmallPZ > iwBigPZ) {
          iwBigNZ = DEF_BIG_SIZE_NZ;
          iwSmallNZ = DEF_SMALL_SIZE_NZ;

          iwBigPZ = DEF_BIG_SIZE_PZ;
          iwSmallPZ = DEF_SMALL_SIZE_PZ;
        }
        if (!fShowTaskbar) {
          fShowNotification = 0;
        }
      }
    }

    in.close();
  } else { // default values
    WelcomeDialog welcome(this, "Welcome to KSmoothDock", TRUE);
    welcome.exec();

    zoomMode = welcome.getZoomMode();
    fShowPager = welcome.getShowPager();
    fShowTaskbar = welcome.getShowTaskbar();
    fShowWinAllDesks = 1;
    fShowClock = welcome.getShowClock();

    iwBigNZ = DEF_BIG_SIZE_NZ;
    iwSmallNZ = DEF_SMALL_SIZE_NZ;

    iwBigPZ = DEF_BIG_SIZE_PZ;
    iwSmallPZ = DEF_SMALL_SIZE_PZ;

    fShowNotification = 0;
    fShowBorders = 1;
    borderColor = QColor("#b1c4de");
    dockOpacity = DEF_DOCK_OPACITY;
    bgColor = QColor("#638abd");
    sepColor = QColor("#b1c4de");
    scrWidth = 0;

    activeDeskColor = QColor("yellow");
    inactiveDeskColor = QColor("white");

    clockTextColor = QColor("white");

    QDir homeDir = QDir::home();
    homeDir.mkdir(".ksmoothdock");
    saveConf();
  }
}

// save configuration

void KSmoothDock::saveConf() {
	saveConf(zoomMode);
}

void KSmoothDock::saveConf(int iZoomMode) {
  QFile out(confFile);
  if (out.open(IO_WriteOnly)) {
    QTextStream out_s(&out);
    out_s << "[KSmoothDock Configuration File]" << endl;

    out_s << "# Zoom Mode: 1 means NORMAL, 2 means PARABOLIC" << endl;
    out_s << "ZoomMode=" << iZoomMode << endl << endl;
    out_s << "# Show Pager: 1 means YES, 0 means NO" << endl;
    out_s << "ShowPager=" << fShowPager << endl << endl;
    out_s << "# Show Taskbar: 1 means YES, 0 means NO" << endl;
    out_s << "ShowTaskbar=" << fShowTaskbar << endl << endl;
    out_s << "ShowWindowsFromAllDesktops=" << fShowWinAllDesks << endl << endl;
    out_s << "# Show Clock: 1 means YES, 0 means NO" << endl;
    out_s << "ShowClock=" << fShowClock << endl << endl;
    out_s << "# Show Notification (when an inactive window's title changed): 1 means YES, 0 means NO" << endl;
    out_s << "ShowNotification=" << fShowNotification << endl << endl;
    out_s << "# Show Borders: 1 means YES, 0 means NO" << endl;
    out_s << "ShowBorders=" << fShowBorders << endl << endl;
    out_s << "# Border color: in the form #RRGGBB" << endl;
    out_s << "BorderColor=" << borderColor.name() << endl << endl;
    out_s << "SmallIconSizeNormalZoom=" << iwSmallNZ << endl << endl;
    out_s << "BigIconSizeNormalZoom=" << iwBigNZ << endl << endl;
    out_s << "SmallIconSizeParabolicZoom=" << iwSmallPZ << endl << endl;
    out_s << "BigIconSizeParabolicZoom=" << iwBigPZ << endl << endl;
    out_s << "# Dock Opacity: 0-100 (0 means transparent, 100 means opaque)" << endl;
    out_s << "DockOpacity=" << dockOpacity << endl << endl;
    out_s << "# Background color: in the form #RRGGBB" << endl;
    out_s << "BackgroundColor=" << bgColor.name() << endl << endl;
    out_s << "# Separator's color: in the form #RRGGBB" << endl;

    out_s << "SeparatorColor=" << sepColor.name() << endl << endl;

    out_s << "# Screen's width (you ONLY need to modify this if you're using xinerama)" << endl;
    out_s << "ScreenWidth=" << scrWidth << endl << endl;
    out_s << "# Color of the text of current desktop number in the pager" << endl;
    out_s << "ActiveDesktopColor=" << activeDeskColor.name() << endl << endl;
    out_s << "# Color of the text of inactive desktop numbers in the pager" << endl;
    out_s << "InactiveDesktopColor=" << inactiveDeskColor.name() << endl << endl;
    out_s << "# Color of the clock text" << endl;
    out_s << "ClockTextColor=" << clockTextColor.name() << endl << endl;

    out.close();
  }
}

// load quick launch menu

void KSmoothDock::loadMenu() {
  QDir homeDir = QDir::home();
  if (homeDir.exists(".ksmoothdock/menu/")) {
    QDir dir(menuPath);

    QStringList fileList = dir.entryList( "*");
    for ( QStringList::Iterator it = fileList.begin(); it != fileList.end(); ++it ) {
      if ( (*it) && ((*it) != ".") && ((*it) != "..") ) {
//        cout << ( *it ) << endl;
        items.append(new Item(menuPath + (*it), iwSmall, iwBig));
      }
    }
  } else {
//    cout << "Dir does not exists" << endl;
    homeDir.mkdir(".ksmoothdock/menu/");

    Item *item1 = new Item("kmenu", "dcop kicker kicker popupKMenu 0", "Start Applications", iwSmall, iwBig);
    item1->saveToFile(menuPath + "1 - KMenu");
    items.append(item1);


    Item *item2 = new Item("desktop", "SHOW_DESKTOP", "Show Desktop", iwSmall, iwBig);
    item2->saveToFile(menuPath + "2 - show desktop");
    items.append(item2);

    Item *item3 = new Item("kfm_home", "kfmclient openProfile filemanagement", "Konqueror - Home Directory", iwSmall, iwBig);
    item3->saveToFile(menuPath + "3 - home");
    items.append(item3);

    Item *item4 = new Item("package_network", "konqueror --profile webbrowsing", "Konqueror - Web Browser", iwSmall, iwBig);
    item4->saveToFile(menuPath + "4 - konqueror");
    items.append(item4);
  }
  numLaunchers = (int)items.count();
}

// add pager (in Parabolic Zooming mode)
// we choose "display" icon as the icon for the pager

void KSmoothDock::addPager() {
  int i;

  for (i = 0; i < numDesks; i++) {
    QString cmd = QString("DESKTOP") + QString::number(i+1);
    QString name = QString("Desktop ") + QString::number(i+1);
    Item *item = new Item("display", cmd, name, iwSmall, iwBig);
    items.append(item);
  }
}

// load running tasks

void KSmoothDock::loadTasks() {
  // show task bar

  if (fShowTaskbar) {
    QValueList<WId>::ConstIterator it;
    for ( it = wm->windows().begin(); it != wm->windows().end(); ++it ) {
      if (wm->hasWId(*it)) {
        addTask(*it);
      }
    }

    connect(wm, SIGNAL(windowAdded(WId)), SLOT(windowAdded(WId)));
    connect(wm, SIGNAL(windowRemoved(WId)), SLOT(windowRemoved(WId)));
    connect(wm, SIGNAL(windowChanged(WId, unsigned int)),
      SLOT(windowChanged(WId, unsigned int)));
    connect(wm, SIGNAL(activeWindowChanged(WId)), SLOT(activeWindowChanged(WId)));
  }
}


// add main popup menu

void KSmoothDock::addMainPopupMenu() {
  popup = new KPopupMenu(this);
  popup->insertItem(i18n("Edit Quick Launch &Menu"), this, SLOT(edit()));
  popup->insertItem(i18n("Edit &Preferences"), this, SLOT(editPref()));
  //popup->insertItem(i18n("&Reload configuration"), this, SLOT(restart()));  
  popup->insertSeparator();
  if (zoomMode == NORMAL_ZOOM)
  	popup->insertItem(i18n("&Switch to Parabolic Zooming Mode"), this, SLOT(switchZoomMode()));
  else
        popup->insertItem(i18n("&Switch to Normal Zooming Mode"), this, SLOT(switchZoomMode()));
  popup->insertSeparator();
	popup->insertItem(i18n("&Hide"), this, SLOT(doHide()));
	popup->insertSeparator();
  popup->insertItem(i18n("&About"), this, SLOT(about()));
	popup->insertSeparator();
  popup->insertItem(i18n("E&xit"), this, SLOT(close()));
}

// switch zoom mode

void KSmoothDock::switchZoomMode() {
  if (zoomMode == NORMAL_ZOOM)
    saveConf(PARABOLIC_ZOOM);
  else
    saveConf(NORMAL_ZOOM);
  KMessageBox::information(this, "You need to restart the program for this change to take effect.");
}

// about dialog

void KSmoothDock::addAboutDlg() {
  aboutDlg = new AboutDialog(this);
}

void KSmoothDock::about() {
  aboutDlg->show();
}

/*
void KSmoothDock::enterTimerTicked() {
  fDrawAsJustEntered = FALSE;
  delete enterTimer;
}

*/

// in case notification is turned on, show Notification

void KSmoothDock::onChangeTimerTicked() {
  onChangeAnimStep = 1 - onChangeAnimStep; // 0 or 1
  repaint();
}

// transparency code

void KSmoothDock::updateBackground(const QPixmap& pix) {
	if (wMode == HIDDEN) {
		updateBackgroundHiddenMode(pix);
	} else {
	  if (zoomMode == NORMAL_ZOOM)
  	  updateBackgroundNZ(pix);
  	else
    	updateBackgroundPZ(pix);
	}
}

void KSmoothDock::updateBackgroundHiddenMode(const QPixmap& pix) {
	bitBlt(&hiddenBg, 0, 0, &pix);
	if (dockOpacity > 0)
		KPixmapEffect::fade(hiddenBg, ((float)dockOpacity)*0.01, bgColor);
	repaint();
}

void KSmoothDock::updateBackgroundNZ(const QPixmap& pix) {
  updateBackground();
}

void KSmoothDock::updateBackgroundPZ(const QPixmap& pix) {
  bitBlt(&topBg, 0, 0, &pix, 0, 0, w1 + 1 + iDist, deltaH); // + iDist to fix the bug: incorrect draw when new window added
	if (wallpapers[curDesk - 1] != NULL) {
    QPixmap *pix2 = wallpapers[curDesk - 1];
    bitBlt(&bottomBgOrg, 0, 0, pix2, x1, dh - h0, w1 + 1, h0);
    bitBlt(&bottomBg, 0, 0, pix2, x1, dh - h0, w1 + 1, h0);

    if (dockOpacity > 0)
      KPixmapEffect::fade(bottomBg, ((float)dockOpacity)*0.01, bgColor);
  }

  repaint();
}

void KSmoothDock::updateBackground() {
	if (wMode == HIDDEN) {
		return;
	}

  if (zoomMode == NORMAL_ZOOM)
    updateBackgroundNZ();
  else
    updateBackgroundPZ();
}

void KSmoothDock::updateBackgroundNZ() {
  if (wallpapers[curDesk - 1] != NULL) {
    bottomBg = QPixmap(w, h);
    bitBlt(&bottomBg, 0, 0, (QPixmap*)wallpapers[curDesk - 1], x, dh - h, w, h);

    if (dockOpacity > 0)
      KPixmapEffect::fade(bottomBg, ((float)dockOpacity)*0.01, bgColor);
  }

  ii_updating_first = 0;
  ii_updating_last = (int)items.count() - 1;
  repaint();

}

void KSmoothDock::updateBackgroundPZ() {
  // cout << "Update bg " << w1 << endl;
  if (wallpapers[curDesk - 1] != NULL) {
    QPixmap *pix = wallpapers[curDesk - 1];
    bitBlt(&bottomBgOrg, 0, 0, pix, x1, dh - h0, w1 + 1, h0);
    bitBlt(&bottomBg, 0, 0, pix, x1, dh - h0, w1 + 1, h0);
    bitBlt(&topBg, 0, 0, pix, x1, dh - h1, w1 + 1 + iDist, deltaH); // + iDist to fix the bug: incorrect draw when new window added

    if (dockOpacity > 0)
      KPixmapEffect::fade(bottomBg, ((float)dockOpacity)*0.01, bgColor);
    repaint();
  }
/*
  ii_updating_first = 0;
  ii_updating_last = (int)items.count() - 1;
  repaint();
*/
}

void KSmoothDock::captureAndUpdateBackground() {
  topBg = QPixmap::grabWindow(qt_xrootwin(), x1, dh - h1, w1, deltaH);
  repaint();
}

/*
void KSmoothDock::updateBackground(const QPixmap& pix) {
  bottomBgOrg = QPixmap(pix);
  bottomBg = QPixmap(pix);
  topBg = QPixmap::grabWindow(qt_xrootwin(), x1, dh - h1, w1, deltaH);
  if (dockOpacity > 0)

    KPixmapEffect::fade(bottomBg, ((float)dockOpacity)*0.01, bgColor);
  repaint();
}
*/

// current desktop changed event handler

void KSmoothDock::currentDesktopChanged(int desktop) {
	curDesk = wm->currentDesktop();
	
	if (wMode == HIDDEN) {
		return;
	}
	
	if (zoomMode == PARABOLIC_ZOOM) {
		setStateBelow();
	}
	
	if (!fShowWinAllDesks) {
		if (fShowTaskbar) {
			int count = (int)items.count();
			
			// remove old taskbar icons
			
			if (zoomMode == NORMAL_ZOOM) { 
				for (int i = count - 1; i >= numLaunchers; i--)
					items.removeLast();
			} else {
				for (int i = count - 1; i >= numLaunchers + numDesks; i--)
					items.removeLast();
			}

			// add new taskbar icons
			
			QValueList<WId>::ConstIterator it;
			for ( it = wm->windows().begin(); it != wm->windows().end(); ++it ) {
				if (wm->hasWId(*it)) {
					addTask(*it);
				}
			}
			doUpdateGeometry();
		}
	}

	//prepareWallpaper();
	//loadWallpaper();
	//updateBackground();
}

// prepare to load the wallpaper

void KSmoothDock::prepareWallpaper() {
	DCOPClient *client = KApplication::dcopClient();
	if (!client->isAttached()) {
		client->attach();
		client->registerAs("KSmoothDock");
	}

	QByteArray data;
	QDataStream args(data, IO_WriteOnly);
	args << 1;
	client->send("kdesktop", "KBackgroundIface", "setExport(int)", data );	
}

// load the wallpaper

void KSmoothDock::loadWallpaper() {
	cout << "Load wallpaper " << curDesk << endl;
  if (wallpapers[curDesk - 1] == NULL) {
    wallpapers[curDesk - 1] = new KSharedPixmap();
    connect(wallpapers[curDesk - 1], SIGNAL(done(bool)), SLOT(wallpaperLoaded(bool)));
	}

  wallpapers[curDesk - 1]->loadFromShared(QString("DESKTOP%1").arg(curDesk));
}

void KSmoothDock::wallpaperLoaded(bool success) {
  if (!success) {
    cout << "Cannot load wallpaper" << endl;
    return;
  }

  cout << "wallpaper loaded" << endl;
  fWallpapersReady[curDesk - 1] = true;

  if (zoomMode == NORMAL_ZOOM) {
    if (wallpapers[curDesk - 1] != NULL) {
      tnWallpapers[curDesk - 1] = new QPixmap(wallpapers[curDesk - 1]->convertToImage().smoothScale(pagerDeskW, pagerDeskH));
    }
  }

  updateBackground();
}

// show desktop

void KSmoothDock::showDesktop() {
  ShowDesktop::the()->toggle();
}

// set kicker position if needed

void KSmoothDock::setKickerPosition() {
	DCOPClient *client = KApplication::dcopClient();
	if (!client->isAttached()) {
		client->attach();
		client->registerAs("KSmoothDock");
	}
	
	QByteArray data, replyData;
	QCString replyType;
	client->call("kicker", "Panel", "panelPosition()", data, replyType, replyData );
	QDataStream reply(replyData, IO_ReadOnly);
  if (replyType == "int") {
    int result;
    reply >> result;
    //cout << "the result is: " << result << endl;
	
		if (result == 3) { // 3 means BOTTOM OF THE SCREEN
			ConfKickerDialog dlg(this, "Setup Kicker", true);
			dlg.exec();		
		}
	}
}

// set state above, i.e. set the dock to be top of the Z order of the window manager
// only need in Parabolic zoom case for transparency effect

void KSmoothDock::setStateAbove() {
  if (zoomMode == PARABOLIC_ZOOM)
    KWin::setState(winId(), NET::KeepAbove);
}

// set state below, i.e. set the dock to be bottom of the Z order of the window manager
// only need in Parabolic zoom case for transparency effect

void KSmoothDock::setStateBelow() {
  if (zoomMode == PARABOLIC_ZOOM)
    KWin::setState(winId(), NET::KeepBelow);
}

// show current window stack, for debugging purpose

void KSmoothDock::showWindowStack() {
  const WId *windows;
  int count;
  int i;

  rootinfo = new NETRootInfo(qt_xdisplay(), 0xffffffff);
  windows = rootinfo->clientListStacking();
  count = rootinfo->clientListStackingCount();

  cout << count << endl;
  for (i = 0; i < count; i++) {
    cout << windows[i] << " ";
    KWin::Info info = KWin::info(windows[i]);
    cout << info.name << " " << info.state << endl;
  }
  delete rootinfo;
}

// minimise dock, used in Parabolic Zooming mode

void KSmoothDock::doMinimise() {
  wMode = MINIMISED;
  repaint();
  setStateBelow();
}

// taskbar functions

void KSmoothDock::minApp() {

  KWin::iconifyWindow(cId, true);
}

void KSmoothDock::maxApp() {
  KWin::setState(cId, NET::MaxVert + NET::MaxHoriz);
}

void KSmoothDock::restApp() {
  KWin::clearState(cId, NET::MaxVert + NET::MaxHoriz);
  KWin::deIconifyWindow(cId, true);
}

void KSmoothDock::closeApp() {
  NETRootInfo app(qt_xdisplay(), NET::CloseWindow);
  app.closeWindowRequest(cId);
}

// Normal mode only

// for smooth zooming animation

void KSmoothDock::timerTicked() {
  int i;
  bool keepTimer = FALSE;


  // cout << "ticked" << endl;

  for (i = 0; i < (int)items.count(); i++) {
    if (iSizeInc[i] != 0) {
      iSize[i] += iSizeInc[i] * (iwBig - iwSmall) / ANIM_TOTAL_STEP;

      if (iSize[i] < iwSmall) {
        iSize[i] = iwSmall;
      }

      if (iSize[i] > iwBig) {
        iSize[i] = iwBig;
      }
    }
  }

  repaint();

  for (i = 0; i < (int)items.count(); i++)
    if (iSizeInc[i] != 0)
      keepTimer = TRUE;

  if (!keepTimer) {
    delete animTimer;
    animTimer = NULL;
    ii_updating_first = 0;
    ii_updating_last = (int)items.count() - 1;
  } else {
    for (i = 0; i < (int)items.count(); i++) {
      if (iSize[i] == iwSmall || iSize[i] == iwBig) {
        iSizeInc[i] = 0;
        if (ii_updating_first == i) {
          if (ii_updating_first < ii_updating_last)
            ii_updating_first = i + 1;
          else
            ii_updating_last = -1;
        }
        if (ii_updating_last == i) {
          if (ii_updating_first < ii_updating_last)
            ii_updating_last = i - 1;
          else
            ii_updating_last = -1;
        }
      }
    }
  }
}

// get index of the current icon, that is, the icon with mouse focus

int KSmoothDock::getIndex(int x) {
  if (x >= 1 && x < cx[numLaunchers - 1] + iwSmall/2 + iSpace/2) {
    return (int)floor((x - ix + iSpace/2)/iDist);
  } else if (x > cx[numLaunchers - 1] + iwSmall/2 + iSpace/2 + pagerW && x <= w - clockW - ix - 1) {
    return (int)floor((x - ix + iSpace/2 - pagerW)/iDist);
  } else if (x >= w - clockW - ix && x < w - ix) { // clock
    return (int)items.count();
  }
  return -1;
}

// stop the effect when the user click on a quick launcher

void KSmoothDock::stopTaskLauncherNotification() {
  iOnClick = -1;
	
	// return to minimised mode
	// the reason we cannot stay in Maximised mode is that: if the newly opened window overlap with the dock, there is NO way that a dock can keep transparency effect (at least I don't know how - Viet)
	
  setStateBelow();
  wMode = MINIMISED;
  repaint();
	
	// we disable ParabolicZoom for 2 seconds, to avoid flicker
	
	QTimer::singleShot( 2000, this, SLOT( reenableParabolicZoom() ) );
}

// reenable parabolic zoom

void KSmoothDock::reenableParabolicZoom() {
	fJustEntered = true;
}

// get and set configuration variables

int KSmoothDock::getZoomMode() {
  return zoomMode;
}

int KSmoothDock::getShowPager() {
  return fShowPager;
}

int KSmoothDock::getShowTaskbar() {
  return fShowTaskbar;
}

int KSmoothDock::getShowWinAllDesks() {
  return fShowWinAllDesks;
}

int KSmoothDock::getShowClock() {
  return fShowClock;
}

int KSmoothDock::getShowNotification() {
  return fShowNotification;
}

int KSmoothDock::getShowBorders() {
  return fShowBorders;
}

QColor KSmoothDock::getBorderColor() {
  return borderColor;
}

int KSmoothDock::getSmallIconSizeNZ() {
  return iwSmallNZ;
}

int KSmoothDock::getBigIconSizeNZ() {
  return iwBigNZ;
}

int KSmoothDock::getSmallIconSizePZ() {
  return iwSmallPZ;
}

int KSmoothDock::getBigIconSizePZ() {
  return iwBigPZ;
}

int KSmoothDock::getDockOpacity() {
  return dockOpacity;
}

QColor KSmoothDock::getBgColor() {
  return bgColor;
}

QColor KSmoothDock::getSepColor() {
  return sepColor;
}

int KSmoothDock::getScrWidth() {
  return scrWidth;
}

QColor KSmoothDock::getActiveDeskColor() {
  return activeDeskColor;
}

QColor KSmoothDock::getInactiveDeskColor() {
  return inactiveDeskColor;
}

QColor KSmoothDock::getClockTextColor() {
  return clockTextColor;
}

void KSmoothDock::setShowPager(int val) {
  fShowPager = val;
}

void KSmoothDock::setShowTaskbar(int val) {
  fShowTaskbar = val;
}

void KSmoothDock::setShowWinAllDesks(int val) {
  fShowWinAllDesks = val;
}

void KSmoothDock::setShowClock(int val) {
  fShowClock = val;
}

void KSmoothDock::setShowNotification(int val) {
  fShowNotification = val;
}

void KSmoothDock::setShowBorders(int val) {
  fShowBorders = val;
}

void KSmoothDock::setBorderColor(QColor val) {
  borderColor = val;
}

void KSmoothDock::setSmallIconSizeNZ(int val) {
  iwSmallNZ = val;
}

void KSmoothDock::setBigIconSizeNZ(int val) {
  iwBigNZ = val;
}

void KSmoothDock::setSmallIconSizePZ(int val) {
  iwSmallPZ = val;
}

void KSmoothDock::setBigIconSizePZ(int val) {
  iwBigPZ = val;
}

void KSmoothDock::setDockOpacity(int val) {
  dockOpacity = val;
}

void KSmoothDock::setBgColor(QColor val) {
  bgColor = val;
}

void KSmoothDock::setSepColor(QColor val) {
  sepColor = val;
}

void KSmoothDock::setScrWidth(int val) {
  scrWidth = val; 
}

void KSmoothDock::setActiveDeskColor(QColor val) {
  activeDeskColor = val;
}

void KSmoothDock::setInactiveDeskColor(QColor val) {
  inactiveDeskColor = val;
}

void KSmoothDock::setClockTextColor(QColor val) {
  clockTextColor = val;
}

// repaint the dock to in case the icon sizes have been changed

void KSmoothDock::updateIconSizes() {
  if (zoomMode == NORMAL_ZOOM) {
    iwSmall = iwSmallNZ;
    iwBig = iwBigNZ;
  } else {
    iwSmall = iwSmallPZ;
    iwBig = iwBigPZ;
  }

  items.clear();
  // load the quick launch menu

  loadMenu();

  // add pager

  if (zoomMode == PARABOLIC_ZOOM) {
    if (fShowPager)
      addPager();
  }

  // load running tasks

  loadTasks();

  doUpdateGeometry();
	
	// update wallpaper thumbnail in NORMAL ZOOM MODE
	
	if (zoomMode == NORMAL_ZOOM && fShowPager) {
		for (int i = 0; i < numDesks; i++) {
    	if (wallpapers[i]) {
				if (tnWallpapers[i])
					delete tnWallpapers[i];
      	tnWallpapers[i] = new QPixmap(wallpapers[i]->convertToImage().smoothScale(pagerDeskW, pagerDeskH));
    	}
		}
  }
	
  repaint();
}

// repaint the dock to in case the components have been changed

void KSmoothDock::updateComponents() {
  // remove all items, except quick launchers

  for (int i = (int)items.count() - 1; i >= numLaunchers; i--)
    items.removeLast();

  // add pager

	if (fShowPager) {
  	if (zoomMode == NORMAL_ZOOM) {    
			for (int i = 0; i < numDesks; i++) {
				if (wallpapers[i] && fWallpapersReady[i]) {
					if (tnWallpapers[i])
						delete tnWallpapers[i];
					tnWallpapers[i] = new QPixmap(wallpapers[i]->convertToImage().smoothScale(pagerDeskW, pagerDeskH));
				}
			}  
		} else { // PARABOLIC ZOOM
			addPager();
		}
  }

  // load running tasks

  loadTasks();
	
	// clock
	
	if (fShowClock) {
		if (!clockTimer) {
			clockTimer = new QTimer(this);
    	connect( clockTimer, SIGNAL(timeout()), this, SLOT(clockTicked()) );
    	clockTimer->start( 60000, FALSE );
    	fUpdateClock = false;
		}
	} else { // if not fShowClock
		if (clockTimer) {
			delete clockTimer;
			clockTimer = NULL;
			fUpdateClock = false;
		}
	}

  // update dock's geometry
  
  doUpdateGeometry();

  repaint();
}

// repaint in case Screen Width has been changed (Xinerama support)

void KSmoothDock::updateScrWidth() {
  if (scrWidth != 0) { // if using xinerama
    dw = scrWidth;
  } else {
    dw = QApplication::desktop()->width();
  }
  doUpdateGeometry();
  repaint();
}

// Hide the dock, works like kicker's hide function. That is, the dock will appear as a small button on the bottom left of the screen

void KSmoothDock::doHide() {
	KWin::setStrut( winId(), 0, 0, 0, 0 );
	wMode = HIDDEN;
	hh = (zoomMode == NORMAL_ZOOM) ? h : h0; // hidden mode height
	hw = HIDDEN_WIDTH;	
	KWin::setState(winId(), NET::KeepAbove | NET::Sticky);
	resize(hw, hh);
	move(0, dh - hh);	
}

// show the dock, i.e. quit the Hidden mode

void KSmoothDock::doShow() {
	toolTipWin->hide();
	prevInd = curInd = -1;
	wMode = MINIMISED;
	fJustEntered = true;
	KWin::clearState(winId(), NET::KeepAbove | NET::Sticky);
	updateComponents();
}
