/***************************************************************************
 *   Copyright (C) 2004 by KSmoothDock team                                *
 *   dangvd@yahoo.com                                                      *
 *   Based on xosd.cpp (C) 2004 by KoolDock team                           * 
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "tooltip.h"

#include <iostream.h>

#include <qpainter.h>
#include <qpixmap.h>
#include <qbitmap.h>
#include <qstring.h>
#include <qfont.h>
#include <qfontmetrics.h>

#include <qwidget.h>
#include <klocale.h>
#include <netwm.h>
#include <netwm_def.h>
#include <kwin.h>

#include <kdebug.h>

ToolTip::ToolTip(QWidget *parent, const char *name) : QWidget(parent, name, WStyle_Customize |  WRepaintNoErase | WStyle_NoBorder | WDestructiveClose | WResizeNoErase | WMouseNoMask | WStyle_StaysOnTop | WX11BypassWM)
{
   info = new NETWinInfo( qt_xdisplay(), winId(), qt_xrootwin(), NET::WMState);
   info->setDesktop( NETWinInfo::OnAllDesktops );
   KWin::setState(winId(), NET::SkipTaskbar | NET::SkipPager | NET::KeepAbove);
}

ToolTip::~ToolTip()
{
}

void ToolTip::setColor(QColor color)
{
   fontColor = color;
}

void ToolTip::setBgColor(QColor color) {
	bgColor = color;
}

void ToolTip::setShadowOffset(int off)
{
   shadowOffset=off;
}

void ToolTip::setText(QString t)
{
   text = t;
   QFontMetrics *metrics = new QFontMetrics(f);
   w = metrics->width(text);
   w = w + shadowOffset + 5; // 5 pixels more
   h = metrics->height();
	
   resize(w,h);	 
   update();
}

void ToolTip::setFontFace(QString fontFace)
{
   f.setFamily(fontFace);
}

void ToolTip::setItalic()
{
   f.setItalic(TRUE);
}

void ToolTip::setBold()
{
   f.setBold(TRUE);
}

void ToolTip::setSize(int size)
{
   f.setPointSize(size);
   fontSize = size;
}

void ToolTip::paintEvent(QPaintEvent *)
{
   QPixmap pm(size());
   QPainter p;
	
   p.begin(&pm, this);
   // fill the painter with background color
   p.fillRect(rect(), bgColor);
   // now we must draw the text with black color for making the mask
   p.setPen( Qt::black );
	 f.setPointSize(fontSize);
   p.setFont(f);
	 
	 if (shadowOffset != 0)
   		p.drawText(shadowOffset, fontSize + shadowOffset, text);	// shadow   
	
   // now, draw normal text
   p.setPen( QColor(fontColor) );	 
	 f.setPointSize(fontSize);
   p.setFont(f);
   p.drawText(1, fontSize, text);	// front	 
	
   p.end();
   bitBlt(this, 0, 0, &pm);	// update the widget
   info->setState(NETWinInfo::SkipTaskbar | NETWinInfo::SkipPager, NETWinInfo::SkipTaskbar | NETWinInfo::SkipPager);
}


